"""Strategies determine which nodes are selected during traversal."""

from .base import NodeTracker, Strategy
from .eager import Eager
from .mmr import Mmr
from .scored import Scored

__all__ = [
    "Eager",
    "Mmr",
    "NodeTracker",
    "Scored",
    "Strategy",
]



import asyncio

import pytest

pytest.register_assert_rewrite("graph_retriever.testing")

from tests.testing.adapters import animals  # noqa: E402
from tests.testing.invoker import sync_or_async  # noqa: E402

# Mark these imports as used so they don't get removed.
# They need to be imported in `conftest.py` so the fixtures are registered.
_ = (
    animals,
    sync_or_async,
)


# Event Loop for async tests.
@pytest.fixture(scope="session")
def event_loop():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    yield loop
    loop.close()



from __future__ import annotations

import dataclasses
from collections.abc import Callable
from typing import Any


@dataclasses.dataclass
class Content:
    """
    Model representing retrieved content.

    Parameters
    ----------
    id :
        The ID of the content.
    content :
        The content.
    embedding :
        The embedding of the content.
    metadata :
        The metadata associated with the content.
    mime_type :
        The MIME type of the content.
    """

    id: str
    content: str
    embedding: list[float]
    metadata: dict[str, Any] = dataclasses.field(default_factory=dict)
    mime_type: str = "text/plain"

    @staticmethod
    def new(
        id: str,
        content: str,
        embedding: list[float] | Callable[[str], list[float]],
        *,
        metadata: dict[str, Any] | None = None,
        mime_type: str = "text/plain",
    ) -> Content:
        """
        Create a new content.

        Parameters
        ----------
        id :
            The ID of the content.
        content :
            The content.
        embedding :
            The embedding, or a function to apply to the content to compute the
            embedding.
        metadata :
            The metadata associated with the content.
        mime_type :
            The MIME type of the content.

        Returns
        -------
        :
            The created content.
        """
        return Content(
            id=id,
            content=content,
            embedding=embedding(content) if callable(embedding) else embedding,
            metadata=metadata or {},
            mime_type=mime_type,
        )



import asyncio
import dataclasses
from collections.abc import AsyncIterator
from typing import TypeVar

T = TypeVar("T")


@dataclasses.dataclass
class _Done:
    exception: bool = False


async def amerge(
    *async_iterables: AsyncIterator[T],
    queue_size: int = 10,
) -> AsyncIterator[T]:
    """
    Merge async iterables into a single async iterator.

    Elements are yielded in the order they become available.

    Parameters
    ----------
    async_iterables :
        The async iterators to merge.
    queue_size :
        Number of elements to buffer in the queue.

    Yields
    ------
    :
        The elements of the iterators as they become available.
    """
    queue: asyncio.Queue[T | _Done] = asyncio.Queue(queue_size)

    async def pump(aiter: AsyncIterator[T]) -> None:
        try:
            async for item in aiter:
                await queue.put(item)
            await queue.put(_Done(exception=False))
        except:
            await queue.put(_Done(exception=True))
            raise

    tasks = [asyncio.create_task(pump(aiter)) for aiter in async_iterables]

    try:
        pending_count = len(async_iterables)
        while pending_count > 0:
            item = await queue.get()
            if isinstance(item, _Done):
                if item.exception:
                    # If there has been an exception, end early.
                    break
                else:
                    pending_count -= 1
            else:
                yield item
            queue.task_done()
    finally:
        for task in tasks:
            if not task.done():
                task.cancel()
        await asyncio.gather(*tasks)



from graph_retriever.utils.batched import batched


def test_batched():
    assert tuple(batched([], 2)) == ()
    assert tuple(batched([0, 1, 2, 3, 4], 2)) == ((0, 1), (2, 3), (4,))
    assert tuple(batched([0, 1, 2, 3, 4, 5], 2)) == ((0, 1), (2, 3), (4, 5))



"""Implements the traversal logic for graph-based document retrieval."""

import copy
from collections.abc import Iterable, Sequence
from typing import Any

from graph_retriever.adapters import Adapter
from graph_retriever.content import Content
from graph_retriever.edges import Edge, EdgeFunction, EdgeSpec, MetadataEdgeFunction
from graph_retriever.strategies import NodeTracker, Strategy
from graph_retriever.types import Node
from graph_retriever.utils.math import cosine_similarity


def traverse(
    query: str,
    *,
    edges: list[EdgeSpec] | EdgeFunction,
    strategy: Strategy,
    store: Adapter,
    metadata_filter: dict[str, Any] | None = None,
    initial_root_ids: Sequence[str] = (),
    store_kwargs: dict[str, Any] = {},
) -> list[Node]:
    """
    Perform a graph traversal to retrieve nodes for a specific query.

    Parameters
    ----------
    query :
        The query string for the traversal.
    edges :
        A list of [EdgeSpec][graph_retriever.edges.EdgeSpec] for use in creating a
        [MetadataEdgeFunction][graph_retriever.edges.MetadataEdgeFunction],
        or an [EdgeFunction][graph_retriever.edges.EdgeFunction].
    strategy :
        The traversal strategy that defines how nodes are discovered, selected,
        and finalized.
    store :
        The vector store adapter used for similarity searches and document
        retrieval.
    metadata_filter :
        Optional filter for metadata during traversal.
    initial_root_ids :
        IDs of the initial root nodes for the traversal.
    store_kwargs :
        Additional arguments passed to the store adapter.

    Returns
    -------
    :
        Nodes returned by the traversal.
    """
    traversal = _Traversal(
        query=query,
        edges=edges,
        strategy=copy.deepcopy(strategy),
        store=store,
        metadata_filter=metadata_filter,
        initial_root_ids=initial_root_ids,
        store_kwargs=store_kwargs,
    )
    return traversal.traverse()


async def atraverse(
    query: str,
    *,
    edges: list[EdgeSpec] | EdgeFunction,
    strategy: Strategy,
    store: Adapter,
    metadata_filter: dict[str, Any] | None = None,
    initial_root_ids: Sequence[str] = (),
    store_kwargs: dict[str, Any] = {},
) -> list[Node]:
    """
    Asynchronously perform a graph traversal to retrieve nodes for a specific query.

    Parameters
    ----------
    query :
        The query string for the traversal.
    edges :
        A list of [EdgeSpec][graph_retriever.edges.EdgeSpec] for use in creating a
        [MetadataEdgeFunction][graph_retriever.edges.MetadataEdgeFunction],
        or an [EdgeFunction][graph_retriever.edges.EdgeFunction].
    strategy :
        The traversal strategy that defines how nodes are discovered, selected,
        and finalized.
    store :
        The vector store adapter used for similarity searches and document
        retrieval.
    metadata_filter :
        Optional filter for metadata during traversal.
    initial_root_ids :
        IDs of the initial root nodes for the traversal.
    store_kwargs :
        Additional arguments passed to the store adapter.

    Returns
    -------
    :
        Nodes returned by the traversal.
    """
    traversal = _Traversal(
        query=query,
        edges=edges,
        strategy=copy.deepcopy(strategy),
        store=store,
        metadata_filter=metadata_filter,
        initial_root_ids=initial_root_ids,
        store_kwargs=store_kwargs,
    )
    return await traversal.atraverse()


class _Traversal:
    """
    Handles a single traversal operation for a graph-based retrieval system.

    The `_Traversal` class manages the process of discovering, visiting, and selecting
    nodes within a graph, based on a query and a traversal strategy. It supports
    synchronous and asynchronous traversal, enabling retrieval of documents in a
    controlled, iterative manner.

    This class should not be reused between traversals.
    """

    def __init__(
        self,
        query: str,
        *,
        edges: list[EdgeSpec] | EdgeFunction,
        strategy: Strategy,
        store: Adapter,
        metadata_filter: dict[str, Any] | None = None,
        initial_root_ids: Sequence[str] = (),
        store_kwargs: dict[str, Any] = {},
    ) -> None:
        self.query = query

        self.edge_function: EdgeFunction
        if isinstance(edges, list):
            self.edge_function = MetadataEdgeFunction(edges)
        elif callable(edges):
            self.edge_function = edges
        else:
            raise ValueError(f"Invalid edges: {edges}")

        self.strategy = strategy
        self.store = store
        self.metadata_filter = metadata_filter
        self.initial_root_ids = initial_root_ids
        self.store_kwargs = store_kwargs

        self._used = False
        self._visited_edges: set[Edge] = set()
        self._edge_depths: dict[Edge, int] = {}
        self._discovered_node_ids: set[str] = set()
        self._node_tracker: NodeTracker = NodeTracker(
            select_k=strategy.select_k, max_depth=strategy.max_depth
        )

    def _check_first_use(self):
        assert not self._used, "Traversals cannot be re-used."
        self._used = True

    def traverse(self) -> list[Node]:
        """
        Execute the traversal synchronously.

        This method retrieves initial candidates, discovers and visits nodes,
        and explores edges iteratively until the traversal is complete.

        Returns
        -------
        :
            The final set of nodes resulting from the traversal.
        """
        self._check_first_use()

        # Retrieve initial candidates.
        initial_content = self._fetch_initial_candidates()
        if self.initial_root_ids:
            initial_content.extend(self.store.get(self.initial_root_ids))
        nodes = self._contents_to_new_nodes(initial_content, depth=0)

        while True:
            self.strategy.iteration(nodes=nodes, tracker=self._node_tracker)

            if self._node_tracker._should_stop_traversal():
                break

            next_outgoing_edges = self.select_next_edges(self._node_tracker.to_traverse)
            new_content = self._fetch_adjacent(next_outgoing_edges)
            nodes = self._contents_to_new_nodes(new_content)

            self._node_tracker.to_traverse.clear()

        return list(self.strategy.finalize_nodes(self._node_tracker.selected))

    async def atraverse(self) -> list[Node]:
        """
        Execute the traversal asynchronously.

        This method retrieves initial candidates, discovers and visits nodes,
        and explores edges iteratively until the traversal is complete.

        Returns
        -------
        :
            The final set of nodes resulting from the traversal.
        """
        self._check_first_use()

        # Retrieve initial candidates.
        initial_content = await self._afetch_initial_candidates()
        if self.initial_root_ids:
            initial_content.extend(await self.store.aget(self.initial_root_ids))
        new_nodes = self._contents_to_new_nodes(initial_content, depth=0)

        while True:
            self.strategy.iteration(nodes=new_nodes, tracker=self._node_tracker)

            if self._node_tracker._should_stop_traversal():
                break

            next_outgoing_edges = self.select_next_edges(self._node_tracker.to_traverse)
            next_content = await self._afetch_adjacent(next_outgoing_edges)
            new_nodes = self._contents_to_new_nodes(next_content)

            self._node_tracker.to_traverse.clear()

        return list(self.strategy.finalize_nodes(self._node_tracker.selected))

    def _fetch_initial_candidates(self) -> list[Content]:
        """
        Retrieve initial candidates based on the query and strategy.

        Returns
        -------
        :
            The initial content retrieved via similarity search.
        """
        query_embedding, docs = self.store.search_with_embedding(
            query=self.query,
            k=self.strategy.start_k,
            filter=self.metadata_filter,
            **self.store_kwargs,
        )
        self.strategy._query_embedding = query_embedding
        return docs

    async def _afetch_initial_candidates(self) -> list[Content]:
        query_embedding, docs = await self.store.asearch_with_embedding(
            query=self.query,
            k=self.strategy.start_k,
            filter=self.metadata_filter,
            **self.store_kwargs,
        )
        self.strategy._query_embedding = query_embedding
        return docs

    def _fetch_adjacent(self, edges: set[Edge]) -> Iterable[Content]:
        """
        Retrieve documents adjacent to the specified outgoing edges.

        This method uses the vector store adapter to fetch documents connected to
        the provided edges.

        Parameters
        ----------
        edges :
            The edges whose adjacent documents need to be fetched.

        Returns
        -------
        :
            The set of content adjacent to the specified edges.
        """
        return self.store.adjacent(
            edges=edges,
            query_embedding=self.strategy._query_embedding,
            k=self.strategy.adjacent_k,
            filter=self.metadata_filter,
            **self.store_kwargs,
        )

    async def _afetch_adjacent(self, edges: set[Edge]) -> Iterable[Content]:
        """
        Asynchronously retrieve documents adjacent to the specified outgoing edges.

        This method uses the vector store adapter to fetch documents connected to
        the provided edges.

        Parameters
        ----------
        edges :
            The edges whose adjacent documents need to be fetched.

        Returns
        -------
        :
            The set of content adjacent to the specified edges.
        """
        return await self.store.aadjacent(
            edges=edges,
            query_embedding=self.strategy._query_embedding,
            k=self.strategy.adjacent_k,
            filter=self.metadata_filter,
            **self.store_kwargs,
        )

    def _contents_to_new_nodes(
        self, contents: Iterable[Content], *, depth: int | None = None
    ) -> Iterable[Node]:
        """
        Convert a content object into a node for traversal.

        This method creates a new `Node` instance, associates it with the document's
        metadata, and calculates its depth based on the incoming edges.

        Parameters
        ----------
        contents :
            The contents to convert into a node.
        depth :
            The depth of the node. If None, the depth is calculated based on the
            incoming edges.

        Returns
        -------
        :
            The newly discovered nodes.
        """
        # Determine which contents to include.

        # TODO: We could push this filtering down into the `adjacent` calls,
        # which could allow some stores to avoid retrieving/decoding already
        # discovered contents. This would complicate the implementation, so we
        # should only do so if the cost of retrieving/decoding is significant.
        content_dict = {
            c.id: c for c in contents if c.id not in self._discovered_node_ids
        }

        if len(content_dict) == 0:
            return []

        # Compute scores.
        scores: list[float] = cosine_similarity(
            [self.strategy._query_embedding],
            [c.embedding for c in content_dict.values()],
        )[0]

        # Create the nodes
        nodes = []
        for content, score in zip(content_dict.values(), scores):
            # Determine incoming/outgoing edges.
            edges = self.edge_function(content)

            # Compute the depth
            if depth is None:
                depth = min(
                    [
                        d
                        for e in edges.incoming
                        if (d := self._edge_depths.get(e, None)) is not None
                    ],
                    default=0,
                )

            nodes.append(
                Node(
                    id=content.id,
                    content=content.content,
                    depth=depth,
                    embedding=content.embedding,
                    similarity_score=score,
                    metadata=content.metadata,
                    incoming_edges=edges.incoming,
                    outgoing_edges=edges.outgoing,
                )
            )
        self._discovered_node_ids.update(content_dict.keys())
        return nodes

    def select_next_edges(self, nodes: dict[str, Node]) -> set[Edge]:
        """
        Find the unvisited outgoing edges from the set of new nodes to traverse.

        This method updates the traversal state by recording the outgoing edges of the
        provided nodes. Outgoing edges that have not been visited before are identified
        and added to the set of edges to explore in subsequent traversal steps.

        Parameters
        ----------
        nodes :
            The new nodes to traverse

        Returns
        -------
        :
            The set of new outgoing edges that need to be explored.

        Notes
        -----
        - The `new_outgoing_edges` dictionary tracks the depth of each outgoing
        edge.
        - If a node's outgoing edge leads to a lower depth, the edge's depth is
        updated to reflect the shortest path.
        - The `_visited_edges` set is updated to include all outgoing edges
        from the provided nodes.
        """
        new_outgoing_edges: dict[Edge, int] = {}
        for node in nodes.values():
            node_new_outgoing_edges = node.outgoing_edges - self._visited_edges
            for edge in node_new_outgoing_edges:
                depth = new_outgoing_edges.setdefault(edge, node.depth + 1)
                if node.depth + 1 < depth:
                    new_outgoing_edges[edge] = node.depth + 1

        self._edge_depths.update(new_outgoing_edges)

        new_outgoing_edge_set = set(new_outgoing_edges.keys())
        self._visited_edges.update(new_outgoing_edge_set)
        return new_outgoing_edge_set



import abc
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, TypeAlias

from immutabledict import immutabledict

from graph_retriever import Content


class Edge(abc.ABC):
    """
    An edge identifies properties necessary for finding matching nodes.

    Sub-classes should be hashable.
    """

    pass


@dataclass(frozen=True)
class MetadataEdge(Edge):
    """
    Link to nodes with specific metadata.

    A `MetadataEdge` connects to nodes with either:

    - `node.metadata[field] == value`
    - `node.metadata[field] CONTAINS value` (if the metadata is a collection).

    Parameters
    ----------
    incoming_field :
        The name of the metadata field storing incoming edges.
    value :
        The value associated with the key for this edge
    """

    def __init__(self, incoming_field: str, value: Any) -> None:
        # `self.field = value` and `setattr(self, "field", value)` -- don't work
        # because of frozen. we need to call `__setattr__` directly (as the
        # default `__init__` would do) to initialize the fields of the frozen
        # dataclass.
        object.__setattr__(self, "incoming_field", incoming_field)

        if isinstance(value, dict):
            value = immutabledict(value)
        object.__setattr__(self, "value", value)

    incoming_field: str
    value: Any


@dataclass(frozen=True)
class IdEdge(Edge):
    """
    An `IdEdge` connects to nodes with `node.id == id`.

    Parameters
    ----------
    id :
        The ID of the node to link to.
    """

    id: str


@dataclass
class Edges:
    """
    Information about the incoming and outgoing edges.

    Parameters
    ----------
    incoming :
        Incoming edges that link to this node.
    outgoing :
        Edges that this node link to. These edges should be defined in terms of
        the *incoming* `Edge` they match. For instance, a link from "mentions"
        to "id" would link to `IdEdge(...)`.
    """

    incoming: set[Edge]
    outgoing: set[Edge]


EdgeFunction: TypeAlias = Callable[[Content], Edges]
"""A function for extracting edges from nodes.

Implementations should be deterministic.
"""



"""Utilities for handling and extracting edges in metadata."""

import warnings
from collections.abc import Iterable
from typing import Any, TypeAlias

from graph_retriever.content import Content

from ._base import Edge, Edges, IdEdge, MetadataEdge

BASIC_TYPES = (str, bool, int, float, complex, bytes)

# Sentinel object used with `dict.get(..., SENTINEL)` calls to distinguish
# between present but `None` (returns `None`) and absent (returns `SENTINEL`)
# elements.
SENTINEL = object()

ID_MAGIC_STRING = "$id"


class Id:
    """
    Place-holder type indicating that the ID should be used.

    Deprecated: Use "$id" instead.
    """

    pass


EdgeSpec: TypeAlias = tuple[str | Id, str | Id]
"""
The definition of an edge for traversal, represented as a pair of fields
representing the source and target of the edge. Each may be:

- A string, `key`, indicating `doc.metadata[key]` as the value.
- The magic string `"$id"`, indicating `doc.id` as the value.

Examples
--------
```
url_to_href_edge          = ("url", "href")
keywords_to_keywords_edge = ("keywords", "keywords")
mentions_to_id_edge       = ("mentions", "$id")
id_to_mentions_edge       = ("$id", "mentions)
```
"""


def _nested_get(metadata: dict[str, Any], key: str) -> Any:
    value = metadata
    for key_part in key.split("."):
        value = value.get(key_part, SENTINEL)
        if value is SENTINEL:
            break
    return value


class MetadataEdgeFunction:
    """
    Helper for extracting and encoding edges in metadata.

    This class provides tools to extract incoming and outgoing edges from
    document metadata. Both incoming and outgoing edges use the same target
    name, enabling equality matching for keys.

    Parameters
    ----------
    edges :
        Definitions of edges for traversal, represented as a pair of fields
        representing the source and target of the edges.

    Raises
    ------
    ValueError
        If an invalid edge definition is provided.
    """

    def __init__(
        self,
        edges: list[EdgeSpec],
    ) -> None:
        self.edges = edges
        for source, target in edges:
            if not isinstance(source, str | Id):
                raise ValueError(f"Expected 'str | Id' but got: {source}")
            if not isinstance(target, str | Id):
                raise ValueError(f"Expected 'str | Id' but got: {target}")

    def _edges_from_dict(
        self,
        id: str,
        metadata: dict[str, Any],
        *,
        incoming: bool = False,
    ) -> set[Edge]:
        """
        Extract edges from the metadata based on declared edge definitions.

        Parameters
        ----------
        metadata :dict[str, Any]
            The metadata dictionary to process.
        incoming :
            If True, extracts edges for incoming relationships.

        Returns
        -------
        :
            A set of edges extracted from the metadata.

        Notes
        -----
        - Handles both simple (key-value) and iterable metadata fields.
        - Issues warnings for unsupported or unexpected values.
        """
        edges: set[Edge] = set()
        for source_key, target_key in self.edges:
            if incoming:
                source_key = target_key

            if target_key == ID_MAGIC_STRING or isinstance(target_key, Id):

                def mk_edge(v) -> Edge:
                    return IdEdge(id=str(v))
            else:

                def mk_edge(v) -> Edge:
                    return MetadataEdge(incoming_field=target_key, value=v)

            if source_key == ID_MAGIC_STRING or isinstance(source_key, Id):
                edges.add(mk_edge(id))
            else:
                value = _nested_get(metadata, source_key)
                if isinstance(value, BASIC_TYPES):
                    edges.add(mk_edge(value))
                elif isinstance(value, Iterable):
                    for item in value:
                        if isinstance(item, BASIC_TYPES):
                            edges.add(mk_edge(item))
                        else:
                            warnings.warn(
                                f"Unsupported item value {item} in '{source_key}'"
                            )
                elif value is not SENTINEL:
                    warnings.warn(f"Unsupported value {value} in '{source_key}'")
        return edges

    def __call__(self, content: Content) -> Edges:
        """
        Extract incoming and outgoing edges for a piece of content.

        This method retrieves edges based on the declared edge definitions, taking
        into account whether nested metadata is used.

        Parameters
        ----------
        content :
            The content to extract edges from.

        Returns
        -------
        :
            the incoming and outgoing edges of the node
        """
        outgoing_edges = self._edges_from_dict(content.id, content.metadata)
        incoming_edges = self._edges_from_dict(
            content.id, content.metadata, incoming=True
        )

        return Edges(incoming=incoming_edges, outgoing=outgoing_edges)



import abc
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

import pytest

from graph_retriever import Content
from graph_retriever.adapters import Adapter
from graph_retriever.edges import Edge, IdEdge, MetadataEdge
from graph_retriever.utils.math import cosine_similarity


def assert_valid_result(content: Content):
    """Assert the content is valid."""
    assert isinstance(content.id, str)
    assert_is_embedding(content.embedding)


def assert_is_embedding(value: Any):
    """Assert the value is an embedding."""
    assert isinstance(value, list)
    for item in value:
        assert isinstance(item, float)


def assert_valid_results(docs: Iterable[Content]):
    """Assert all of the contents are valid results."""
    for doc in docs:
        assert_valid_result(doc)


def assert_ids_any_order(
    results: Iterable[Content],
    expected: list[str],
) -> None:
    """Assert the results are valid and match the IDs."""
    assert_valid_results(results)

    result_ids = [r.id for r in results]
    assert set(result_ids) == set(expected), "should contain exactly expected IDs"


def cosine_similarity_scores(
    adapter: Adapter, query_or_embedding: str | list[float], ids: list[str]
) -> dict[str, float]:
    """Return the cosine similarity scores for the given IDs and query embedding."""
    if len(ids) == 0:
        return {}

    docs = adapter.get(ids)
    found_ids = (d.id for d in docs)
    assert set(ids) == set(found_ids), "can't find all IDs"

    if isinstance(query_or_embedding, str):
        query_embedding = adapter.search_with_embedding(query_or_embedding, k=0)[0]
    else:
        query_embedding = query_or_embedding

    scores: list[float] = cosine_similarity(
        [query_embedding],
        [d.embedding for d in docs],
    )[0]

    return {doc.id: score for doc, score in zip(docs, scores)}


def assert_ids_in_cosine_similarity_order(
    results: Iterable[Content],
    expected: list[str],
    query_embedding: list[float],
    adapter: Adapter,
) -> None:
    """Assert the results are valid and in cosine similarity order."""
    assert_valid_results(results)
    result_ids = [r.id for r in results]

    similarity_scores = cosine_similarity_scores(adapter, query_embedding, expected)
    expected = sorted(expected, key=lambda id: similarity_scores[id], reverse=True)

    assert result_ids == expected, (
        "should contain expected IDs in cosine similarity order"
    )


@dataclass(kw_only=True)
class AdapterComplianceCase(abc.ABC):
    """
    Base dataclass for test cases.

    Attributes
    ----------
    id :
        The ID of the test case.

    expected :
        The expected results of the case.
    """

    id: str
    expected: list[str]

    requires_nested: bool = False
    requires_dict_in_list: bool = False


@dataclass
class GetCase(AdapterComplianceCase):
    """A test case for `get` and `aget`."""

    request: list[str]
    filter: dict[str, Any] | None = None


GET_CASES: list[GetCase] = [
    # Currently, this is not required for `get` implementations since the
    # traversal skips making `get` calls with no IDs. Some stores (such as chroma)
    # fail in this case.
    # GetCase("none", [], []),
    GetCase(id="one", request=["boar"], expected=["boar"]),
    GetCase(
        id="many",
        request=[
            "alligator",
            "barracuda",
            "chameleon",
            "cobra",
            "crocodile",
            "dolphin",
            "eel",
            "fish",
            "gecko",
            "iguana",
            "jellyfish",
            "komodo dragon",
            "lizard",
            "manatee",
            "narwhal",
        ],
        expected=[
            "alligator",
            "barracuda",
            "chameleon",
            "cobra",
            "crocodile",
            "dolphin",
            "eel",
            "fish",
            "gecko",
            "iguana",
            "jellyfish",
            "komodo dragon",
            "lizard",
            "manatee",
            "narwhal",
        ],
    ),
    GetCase(
        id="missing",
        request=["boar", "chinchilla", "unicorn", "cobra"],
        expected=["boar", "chinchilla", "cobra"],
    ),
    GetCase(
        id="duplicate",
        request=["boar", "chinchilla", "boar", "cobra"],
        expected=["boar", "chinchilla", "cobra"],
    ),
    GetCase(
        id="filtered",
        request=["boar", "chinchilla", "boar", "cobra"],
        expected=["chinchilla"],
        filter={"keywords": "andes"},
    ),
]


@dataclass
class SearchCase(AdapterComplianceCase):
    """A test case for `similarity_search_*` and `asimilarity_search_*` methods."""

    query: str
    k: int | None = None
    filter: dict[str, str] | None = None

    @property
    def kwargs(self):
        """Return keyword arguments for the test invocation."""
        kwargs = {}
        if self.k is not None:
            kwargs["k"] = self.k
        if self.filter is not None:
            kwargs["filter"] = self.filter
        return kwargs


SEARCH_CASES: list[SearchCase] = [
    SearchCase(
        id="basic",
        query="domesticated hunters",
        expected=["cat", "horse", "chicken", "dog"],
    ),
    SearchCase(id="k2", query="domesticated hunters", k=2, expected=["cat", "horse"]),
    SearchCase(
        id="k0",
        query="domesticated hunters",
        k=0,
        expected=[],
    ),
    SearchCase(
        id="value_filter",
        query="domesticated hunters",
        filter={"type": "mammal"},
        expected=["cat", "dog", "horse", "alpaca"],
    ),
    SearchCase(
        id="list_filter",
        query="domesticated hunters",
        filter={"keywords": "hunting"},
        expected=["cat"],
    ),
    SearchCase(
        id="two_filters",
        query="domesticated hunters",
        filter={"type": "mammal", "diet": "carnivorous"},
        expected=["cat", "dingo", "ferret"],
    ),
    # OpenSearch supports filtering on multiple values, but it is not currently
    # relied on. Since no other adapters support it, we don't test it nor should
    # traversal depend on it.
    # SimilaritySearchCase(
    #   id="multi_list_filter",
    #   query="domesticated hunters",
    #   filter={"keywords": ["hunting", "agile"]},
    #   expected=["cat", "fox", "gazelle", "mongoose"]
    # ),
]


@dataclass
class AdjacentCase(AdapterComplianceCase):
    """A test case for `get_adjacent` and `aget_adjacent`."""

    query: str
    edges: set[Edge]

    k: int = 4
    filter: dict[str, Any] | None = None


ADJACENT_CASES: list[AdjacentCase] = [
    AdjacentCase(
        id="one_edge",
        query="domesticated hunters",
        edges={MetadataEdge("type", "mammal")},
        expected=["horse", "alpaca", "dog", "cat"],
    ),
    AdjacentCase(
        id="two_edges_same_field",
        query="domesticated hunters",
        edges={
            MetadataEdge("type", "mammal"),
            MetadataEdge("type", "crustacean"),
        },
        expected=[
            "alpaca",
            "cat",
            "dog",
            "horse",
        ],
    ),
    AdjacentCase(
        id="numeric",
        query="domesticated hunters",
        edges={
            MetadataEdge("number_of_legs", 0),
        },
        k=20,  # more than match the filter so we get all
        expected=[
            "barracuda",
            "cobra",
            "dolphin",
            "eel",
            "fish",
            "jellyfish",
            "manatee",
            "narwhal",
        ],
    ),
    AdjacentCase(
        id="two_edges_diff_field",
        query="domesticated hunters",
        edges={
            MetadataEdge("type", "reptile"),
            MetadataEdge("number_of_legs", 0),
        },
        k=20,  # more than match the filter so we get all
        expected=[
            "alligator",
            "barracuda",
            "chameleon",
            "cobra",
            "crocodile",
            "dolphin",
            "eel",
            "fish",
            "gecko",
            "iguana",
            "jellyfish",
            "komodo dragon",
            "lizard",
            "manatee",
            "narwhal",
        ],
    ),
    AdjacentCase(
        id="one_ids",
        query="domesticated hunters",
        edges={
            IdEdge("cat"),
        },
        expected=[
            "cat",
        ],
    ),
    AdjacentCase(
        id="many_ids",
        query="domesticated hunters",
        edges={
            IdEdge("cat"),
            IdEdge("dog"),
            IdEdge("unicorn"),
            IdEdge("crab"),
        },
        expected=[
            "cat",
            "dog",
            "crab",
        ],
    ),
    AdjacentCase(
        id="ids_limit_k",
        query="domesticated hunters",
        edges={
            IdEdge("cat"),
            IdEdge("dog"),
            IdEdge("unicorn"),
            IdEdge("antelope"),
        },
        k=2,
        expected=[
            "cat",
            "dog",
        ],
    ),
    AdjacentCase(
        id="filtered_ids",
        query="domesticated hunters",
        edges={
            IdEdge("boar"),
            IdEdge("chinchilla"),
            IdEdge("unicorn"),
            IdEdge("griaffe"),
        },
        filter={"keywords": "andes"},
        expected=[
            "chinchilla",
        ],
    ),
    AdjacentCase(
        id="metadata_and_id",
        query="domesticated hunters",
        edges={
            IdEdge("cat"),
            MetadataEdge("type", "reptile"),
        },
        k=6,
        expected=[
            "alligator",  # reptile
            "crocodile",  # reptile
            "cat",  # by ID
            "chameleon",  # reptile
            "gecko",  # reptile
            "komodo dragon",  # reptile
        ],
    ),
    AdjacentCase(
        id="dict_in_list",
        query="domesticated hunters",
        edges={
            MetadataEdge("tags", {"a": 5, "b": 7}),
        },
        expected=[
            "aardvark",
        ],
        requires_dict_in_list=True,
    ),
    AdjacentCase(
        id="dict_in_list_multiple",
        query="domesticated hunters",
        edges={
            MetadataEdge("tags", {"a": 5, "b": 7}),
            MetadataEdge("tags", {"a": 5, "b": 8}),
        },
        expected=[
            "aardvark",
            "albatross",
        ],
        requires_dict_in_list=True,
    ),
    AdjacentCase(
        id="absent_dict",
        query="domesticated hunters",
        edges={
            MetadataEdge("tags", {"a": 5, "b": 10}),
        },
        expected=[],
        requires_dict_in_list=True,
    ),
    AdjacentCase(
        id="nested",
        query="domesticated hunters",
        edges={
            MetadataEdge("nested.a", 5),
        },
        expected=[
            "alligator",
            "alpaca",
        ],
        requires_nested=True,
    ),
    AdjacentCase(
        id="nested_same_field",
        query="domesticated hunters",
        edges={
            MetadataEdge("nested.a", 5),
            MetadataEdge("nested.a", 6),
        },
        expected=[
            "alligator",
            "alpaca",
            "ant",
        ],
        requires_nested=True,
    ),
    AdjacentCase(
        id="nested_diff_field",
        query="domesticated hunters",
        edges={
            MetadataEdge("nested.a", 5),
            MetadataEdge("nested.b", 5),
        },
        expected=[
            "alligator",
            "alpaca",
            "anteater",
        ],
        requires_nested=True,
    ),
]


class AdapterComplianceSuite(abc.ABC):
    """
    Test suite for adapter compliance.

    To use this, create a sub-class containing a `@pytest.fixture` named
    `adapter` which returns an `Adapter` with the documents from `animals.jsonl`
    loaded.
    """

    def supports_nested_metadata(self) -> bool:
        """Return whether nested metadata is expected to work."""
        return True

    def supports_dict_in_list(self) -> bool:
        """Return whether dicts can appear in list fields in metadata."""
        return True

    def expected(self, method: str, case: AdapterComplianceCase) -> list[str]:
        """
        Override to change the expected behavior of a case.

        If the test is expected to fail, call `pytest.xfail(reason)`, or
        `pytest.skip(reason)` if it can't be executed.

        Generally, this should *not* change the expected results, unless the the
        adapter being tested uses wildly different distance metrics or a
        different embedding. The `AnimalsEmbedding` is deterministic and the
        results across vector stores should generally be deterministic and
        consistent.

        Parameters
        ----------
        method :
            The method being tested. For instance, `get`, `aget`, or
            `similarity_search_with_embedding`, etc.
        case :
            The case being tested.

        Returns
        -------
        :
            The expected animals.
        """
        if not self.supports_nested_metadata() and case.requires_nested:
            pytest.xfail("nested metadata not supported")
        if not self.supports_dict_in_list() and case.requires_dict_in_list:
            pytest.xfail("dict-in-list fields is not supported")
        return case.expected

    @pytest.fixture(params=GET_CASES, ids=lambda c: c.id)
    def get_case(self, request) -> GetCase:
        """Fixture providing the `get` and `aget` test cases."""
        return request.param

    @pytest.fixture(params=ADJACENT_CASES, ids=lambda c: c.id)
    def adjacent_case(self, request) -> AdjacentCase:
        """Fixture providing the `get_adjacent` and `aget_adjacent` test cases."""
        return request.param

    @pytest.fixture(params=SEARCH_CASES, ids=lambda c: c.id)
    def search_case(self, request) -> SearchCase:
        """Fixture providing the `(a)?similarity_search_*` test cases."""
        return request.param

    def test_get(self, adapter: Adapter, get_case: GetCase) -> None:
        """Run tests for `get`."""
        expected = self.expected("get", get_case)
        results = adapter.get(get_case.request, filter=get_case.filter)
        assert_ids_any_order(results, expected)

    async def test_aget(self, adapter: Adapter, get_case: GetCase) -> None:
        """Run tests for `aget`."""
        expected = self.expected("aget", get_case)
        results = await adapter.aget(get_case.request, filter=get_case.filter)
        assert_ids_any_order(results, expected)

    def test_search_with_embedding(
        self, adapter: Adapter, search_case: SearchCase
    ) -> None:
        """Run tests for `search_with_embedding`."""
        expected = self.expected("search_with_embedding", search_case)
        embedding, results = adapter.search_with_embedding(
            search_case.query, **search_case.kwargs
        )
        assert_is_embedding(embedding)
        assert_ids_in_cosine_similarity_order(results, expected, embedding, adapter)

    async def test_asearch_with_embedding(
        self, adapter: Adapter, search_case: SearchCase
    ) -> None:
        """Run tests for `asearch_with_embedding`."""
        expected = self.expected("asearch_with_embedding", search_case)
        embedding, results = await adapter.asearch_with_embedding(
            search_case.query, **search_case.kwargs
        )
        assert_is_embedding(embedding)
        assert_ids_in_cosine_similarity_order(results, expected, embedding, adapter)

    def test_search(self, adapter: Adapter, search_case: SearchCase) -> None:
        """Run tests for `search`."""
        expected = self.expected("search", search_case)
        embedding, _ = adapter.search_with_embedding(search_case.query, k=0)
        results = adapter.search(embedding, **search_case.kwargs)
        assert_ids_in_cosine_similarity_order(results, expected, embedding, adapter)

    async def test_asearch(self, adapter: Adapter, search_case: SearchCase) -> None:
        """Run tests for `asearch`."""
        expected = self.expected("asearch", search_case)
        embedding, _ = await adapter.asearch_with_embedding(search_case.query, k=0)
        results = await adapter.asearch(embedding, **search_case.kwargs)
        assert_ids_in_cosine_similarity_order(results, expected, embedding, adapter)

    def test_adjacent(self, adapter: Adapter, adjacent_case: AdjacentCase) -> None:
        """Run tests for `adjacent."""
        expected = self.expected("adjacent", adjacent_case)
        embedding, _ = adapter.search_with_embedding(adjacent_case.query, k=0)
        results = adapter.adjacent(
            edges=adjacent_case.edges,
            query_embedding=embedding,
            k=adjacent_case.k,
            filter=adjacent_case.filter,
        )
        assert_ids_in_cosine_similarity_order(results, expected, embedding, adapter)

    async def test_aadjacent(
        self, adapter: Adapter, adjacent_case: AdjacentCase
    ) -> None:
        """Run tests for `aadjacent."""
        expected = self.expected("aadjacent", adjacent_case)
        embedding, _ = await adapter.asearch_with_embedding(adjacent_case.query, k=0)
        results = await adapter.aadjacent(
            edges=adjacent_case.edges,
            query_embedding=embedding,
            k=adjacent_case.k,
            filter=adjacent_case.filter,
        )
        assert_ids_in_cosine_similarity_order(results, expected, embedding, adapter)



import asyncio
from collections.abc import Callable
from concurrent.futures import Executor
from contextvars import copy_context
from functools import partial
from typing import ParamSpec, TypeVar, cast

P = ParamSpec("P")
T = TypeVar("T")


async def run_in_executor(
    executor: Executor | None,
    func: Callable[P, T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:  # noqa: DOC502
    """
    Run a function in an executor.

    Parameters
    ----------
    executor :
        The executor to run in.
    func :
        The function.
    *args :
        The positional arguments to the function.
    kwargs :
        The keyword arguments to the function.

    Returns
    -------
    :
        The output of the function.

    Raises
    ------
    RuntimeError
        If the function raises a StopIteration.
    """  # noqa: DOC502

    def wrapper() -> T:
        try:
            return func(*args, **kwargs)
        except StopIteration as exc:
            # StopIteration can't be set on an asyncio.Future
            # it raises a TypeError and leaves the Future pending forever
            # so we need to convert it to a RuntimeError
            raise RuntimeError from exc

    if executor is None or isinstance(executor, dict):
        # Use default executor with context copied from current context
        return await asyncio.get_running_loop().run_in_executor(
            None,
            cast(Callable[..., T], partial(copy_context().run, wrapper)),
        )

    return await asyncio.get_running_loop().run_in_executor(executor, wrapper)



import dataclasses
import heapq
from collections.abc import Callable, Iterable

from typing_extensions import override

from graph_retriever.strategies.base import NodeTracker, Strategy
from graph_retriever.types import Node


class _ScoredNode:
    def __init__(self, score: float, node: Node) -> None:
        self.score = score
        self.node = node

    def __lt__(self, other: "_ScoredNode") -> bool:
        return other.score < self.score


@dataclasses.dataclass
class Scored(Strategy):
    """
    Scored traversal strategy.

    This strategy uses a scoring function to select nodes using a local maximum
    approach. In each iteration, it chooses the top scoring nodes available and
    then traverses the connected nodes.

    Parameters
    ----------
    scorer:
        A callable function that returns the score of a node.
    select_k :
        Maximum number of nodes to retrieve during traversal.
    start_k :
        Number of documents to fetch via similarity for starting the traversal.
        Added to any initial roots provided to the traversal.
    adjacent_k :
        Number of documents to fetch for each outgoing edge.
    max_depth :
        Maximum traversal depth. If `None`, there is no limit.
    per_iteration_limit:
        Maximum number of nodes to select and traverse during a single
        iteration.
    k:
        Deprecated: Use `select_k` instead.
        Maximum number of nodes to select and return during traversal.
    """

    scorer: Callable[[Node], float]
    _nodes: list[_ScoredNode] = dataclasses.field(default_factory=list)

    per_iteration_limit: int | None = None

    @override
    def iteration(self, nodes: Iterable[Node], tracker: NodeTracker) -> None:
        for node in nodes:
            heapq.heappush(self._nodes, _ScoredNode(self.scorer(node), node))

        limit = tracker.num_remaining
        if self.per_iteration_limit:
            limit = min(limit, self.per_iteration_limit)

        while limit > 0 and self._nodes:
            highest = heapq.heappop(self._nodes)
            node = highest.node
            node.extra_metadata["_score"] = highest.score
            limit -= tracker.select_and_traverse([node])

    @override
    def finalize_nodes(self, selected):
        selected = sorted(
            selected, key=lambda node: node.extra_metadata["_score"], reverse=True
        )
        return super().finalize_nodes(selected)



from collections.abc import Iterable

try:
    # Try importing the function from itertools (Python 3.12+)
    from itertools import batched  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover
    from collections.abc import Iterator
    from itertools import islice
    from typing import TypeVar

    # Fallback implementation for older Python versions

    T = TypeVar("T")

    # This is equivalent to `itertools.batched`, but that is only available in 3.12
    def batched(iterable: Iterable[T], n: int) -> Iterator[tuple[T, ...]]:  # type: ignore[no-redef]
        """
        Equivalent of itertools.batched for pre 3.12.

        Parameters
        ----------
        iterable :
            Iterator over elements.
        n :
            Size of batches.

        Yields
        ------
        Iterator[T]
            Yields iterators over elements in successive batches.

        Raises
        ------
        ValueError
            If `n` is less than 1.
        """
        if n < 1:
            raise ValueError("n must be at least one")

        iterator = iter(iterable)
        while batch := tuple(islice(iterator, n)):
            yield batch



import abc
from collections.abc import Callable, Iterable, Sequence
from typing import Any, TypeAlias

from typing_extensions import override

from graph_retriever.adapters.base import Adapter
from graph_retriever.content import Content
from graph_retriever.utils.math import cosine_similarity

SENTINEL = object()

Embedding: TypeAlias = Callable[[str], list[float]]


class InMemoryBase(Adapter, abc.ABC):
    """
    The base class for in-memory adapters that use dict-based metadata filters.

    These are intended (mostly) for demonstration purposes and testing.
    """

    def __init__(self, embedding: Embedding, content: list[Content]) -> None:
        """
        Initialize with the given embedding function.

        Parameters
        ----------
        embedding :
            embedding function to use.
        """
        self.store: dict[str, Content] = {c.id: c for c in content}
        self.embedding = embedding

    @override
    def search_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[list[float], list[Content]]:
        query_embedding = self.embedding(query)
        docs = self.search(
            embedding=query_embedding,
            k=k,
            filter=filter,
            **kwargs,
        )
        return query_embedding, docs

    @override
    async def asearch_with_embedding(
        self,
        query: str,
        k: int = 4,
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[list[float], list[Content]]:
        query_embedding = self.embedding(query)
        docs = await self.asearch(
            embedding=query_embedding,
            k=k,
            filter=filter,
            **kwargs,
        )
        return query_embedding, docs

    @override
    def search(
        self,
        embedding: list[float],
        k: int = 4,
        filter: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> list[Content]:
        # get all docs with fixed order in list
        candidates = self._matching_content(filter)

        if not candidates:
            return []

        similarity = cosine_similarity([embedding], [c.embedding for c in candidates])[
            0
        ]

        # get the indices ordered by similarity score
        top_k_idx = similarity.argsort()[::-1][:k]

        return [candidates[idx] for idx in top_k_idx]

    @override
    def get(
        self, ids: Sequence[str], filter: dict[str, Any] | None = None, **kwargs: Any
    ) -> list[Content]:
        return [
            c
            for id in ids
            if (c := self.store.get(id, None))
            if self._matches(filter, c)
        ]

    def _matching_content(self, filter: dict[str, Any] | None = None) -> list[Content]:
        """Return a list of content matching the given filters."""
        if filter:
            return [c for c in self.store.values() if self._matches(filter, c)]
        else:
            return list(self.store.values())

    def _matches(self, filter: dict[str, Any] | None, content: Content) -> bool:
        """Return whether `content` matches the given `filter`."""
        if not filter:
            return True

        for key, filter_value in filter.items():
            content_value = content.metadata
            for key_part in key.split("."):
                content_value = content_value.get(key_part, SENTINEL)
                if content_value is SENTINEL:
                    break
            if not self._value_matches(filter_value, content_value):
                return False
        return True

    @abc.abstractmethod
    def _value_matches(self, filter_value: str, content_value: Any) -> bool:
        """Return whether the `content_value` matches the `filter_value`."""
        ...


class InMemory(InMemoryBase):
    """
    In-Memory VectorStore that supports list-based metadata.

    This In-Memory store simulates VectorStores like AstraDB and OpenSearch
    """

    @override
    def _value_matches(self, filter_value: str, content_value: Any) -> bool:
        return (filter_value == content_value) or (
            isinstance(content_value, Iterable)
            and not isinstance(content_value, str | bytes)
            and filter_value in content_value
        )



"""Define the base traversal strategy."""

from __future__ import annotations

import abc
import dataclasses
from collections.abc import Iterable
from typing import Any

from graph_retriever.types import Node

DEFAULT_SELECT_K = 5


class NodeTracker:
    """
    Helper class initiating node selection and traversal.

    Call .select(nodes) to add nodes to the result set.
    Call .traverse(nodes) to add nodes to the next traversal.
    Call .select_and_traverse(nodes) to add nodes to the result set and the next
        traversal.
    """

    def __init__(self, select_k: int, max_depth: int | None) -> None:
        self._select_k: int = select_k
        self._max_depth: int | None = max_depth
        self._visited_node_ids: set[str] = set()
        # use a dict to preserve order
        self.to_traverse: dict[str, Node] = dict()
        self.selected: list[Node] = []

    @property
    def num_remaining(self):
        """The remaining number of nodes to be selected."""
        return max(self._select_k - len(self.selected), 0)

    def select(self, nodes: Iterable[Node]) -> None:
        """Select nodes to be included in the result set."""
        for node in nodes:
            node.extra_metadata["_depth"] = node.depth
            node.extra_metadata["_similarity_score"] = node.similarity_score
        self.selected.extend(nodes)

    def traverse(self, nodes: Iterable[Node]) -> int:
        """
        Select nodes to be included in the next traversal.

        Returns
        -------
        Number of nodes added for traversal.

        Notes
        -----
        - Nodes are only added if they have not been visited before.
        - Nodes are only added if they do not exceed the maximum depth.
        - If no new nodes are chosen for traversal, or selected for output, then
            the traversal will stop.
        - Traversal will also stop if the number of selected nodes reaches the select_k
            limit.
        """
        new_nodes = {
            n.id: n
            for n in nodes
            if self._not_visited(n.id)
            if self._max_depth is None or n.depth < self._max_depth
        }
        self.to_traverse.update(new_nodes)
        self._visited_node_ids.update(new_nodes.keys())
        return len(new_nodes)

    def select_and_traverse(self, nodes: Iterable[Node]) -> int:
        """
        Select nodes to be included in the result set and the next traversal.

        Returns
        -------
        Number of nodes added for traversal.

        Notes
        -----
        - Nodes are only added for traversal if they have not been visited before.
        - Nodes are only added for traversal if they do not exceed the maximum depth.
        - If no new nodes are chosen for traversal, or selected for output, then
            the traversal will stop.
        - Traversal will also stop if the number of selected nodes reaches the select_k
            limit.
        """
        self.select(nodes)
        return self.traverse(nodes)

    def _not_visited(self, id: str):
        """Return true if the node has not been visited."""
        return id not in self._visited_node_ids

    def _should_stop_traversal(self):
        """Return true if traversal should be stopped."""
        return self.num_remaining == 0 or len(self.to_traverse) == 0


@dataclasses.dataclass(kw_only=True)
class Strategy(abc.ABC):
    """
    Interface for configuring node selection and traversal strategies.

    This base class defines how nodes are selected, traversed, and finalized during
    a graph traversal. Implementations can customize behaviors like limiting the depth
    of traversal, scoring nodes, or selecting the next set of nodes for exploration.

    Parameters
    ----------
    select_k :
        Maximum number of nodes to select and return during traversal.
    start_k :
        Number of nodes to fetch via similarity for starting the traversal.
        Added to any initial roots provided to the traversal.
    adjacent_k :
        Number of nodes to fetch for each outgoing edge.
    max_traverse :
        Maximum number of nodes to traverse outgoing edges from before returning.
        If `None`, there is no limit.
    max_depth :
        Maximum traversal depth. If `None`, there is no limit.
    k:
        Deprecated: Use `select_k` instead.
        Maximum number of nodes to select and return during traversal.
    """

    select_k: int = dataclasses.field(default=DEFAULT_SELECT_K)
    start_k: int = 4
    adjacent_k: int = 10
    max_traverse: int | None = None
    max_depth: int | None = None
    k: int = dataclasses.field(default=DEFAULT_SELECT_K, repr=False)

    _query_embedding: list[float] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        """Allow passing the deprecated 'k' value instead of 'select_k'."""
        if self.select_k == DEFAULT_SELECT_K and self.k != DEFAULT_SELECT_K:
            self.select_k = self.k
        else:
            self.k = self.select_k

    @abc.abstractmethod
    def iteration(self, *, nodes: Iterable[Node], tracker: NodeTracker) -> None:
        """
        Process the newly discovered nodes on each iteration.

        This method should call `tracker.traverse()` and/or `tracker.select()`
        as appropriate to update the nodes that need to be traversed in this iteration
        or selected at the end of the retrieval, respectively.

        Parameters
        ----------
        nodes :
            The newly discovered nodes found from either:
            - the initial vector store retrieval
            - incoming edges from nodes chosen for traversal in the previous iteration
        tracker :
            The tracker object to manage the traversal and selection of nodes.

        Notes
        -----
        - This method is called once for each iteration of the traversal.
        - In order to stop iterating either choose to not traverse any additional nodes
        or don't select any additional nodes for output.
        """
        ...

    def finalize_nodes(self, selected: Iterable[Node]) -> Iterable[Node]:
        """
        Finalize the selected nodes.

        This method is called before returning the final set of nodes. It allows
        the strategy to perform any final processing or re-ranking of the selected
        nodes.

        Parameters
        ----------
        selected :
            The selected nodes to be finalized

        Returns
        -------
        :
            Finalized nodes.

        Notes
        -----
        - The default implementation returns the first `self.select_k` selected nodes
        without any additional processing.
        """
        return list(selected)[: self.select_k]

    @staticmethod
    def build(
        base_strategy: Strategy,
        **kwargs: Any,
    ) -> Strategy:
        """
        Build a strategy for a retrieval operation.

        Combines a base strategy with any provided keyword arguments to
        create a customized traversal strategy.

        Parameters
        ----------
        base_strategy :
            The base strategy to start with.
        kwargs :
            Additional configuration options for the strategy.

        Returns
        -------
        :
            A configured strategy instance.

        Raises
        ------
        ValueError
            If 'strategy' is set incorrectly or extra arguments are invalid.
        """
        # Check if there is a new strategy to use. Otherwise, use the base.
        strategy: Strategy
        if "strategy" in kwargs:
            if next(iter(kwargs.keys())) != "strategy":
                raise ValueError("Error: 'strategy' must be set before other args.")
            strategy = kwargs.pop("strategy")
            if not isinstance(strategy, Strategy):
                raise ValueError(
                    f"Unsupported 'strategy' type {type(strategy).__name__}."
                    " Must be a sub-class of Strategy"
                )
        elif base_strategy is not None:
            strategy = base_strategy
        else:
            raise ValueError("'strategy' must be set in `__init__` or invocation")

        # Apply the kwargs to update the strategy.
        assert strategy is not None
        if "k" in kwargs:
            kwargs["select_k"] = kwargs.pop("k")
        strategy = dataclasses.replace(strategy, **kwargs)

        return strategy



import pytest
from graph_retriever.adapters import Adapter
from graph_retriever.adapters.in_memory import InMemory
from graph_retriever.testing.adapter_tests import AdapterComplianceSuite


class TestInMemory(AdapterComplianceSuite):
    @pytest.fixture(scope="class")
    def adapter(self, animals: Adapter) -> Adapter:
        assert isinstance(animals, InMemory)
        return animals



from __future__ import annotations

import abc
import dataclasses
from collections.abc import Callable, Sequence
from typing import Any, Generic, TypeVar

import pytest
from graph_retriever import Node, atraverse, traverse
from graph_retriever.adapters import Adapter
from graph_retriever.edges import EdgeFunction, EdgeSpec
from graph_retriever.strategies import Strategy


class SyncOrAsync(abc.ABC):
    @abc.abstractmethod
    async def _traverse(self, **kwargs: Any) -> list[Node]: ...

    def traverse_sorted_ids(
        self,
        store: Adapter,
        query: str | None = None,
        edges: list[EdgeSpec] | EdgeFunction | None = None,
        strategy: Strategy | None = None,
    ) -> TraversalCall[list[str]]:
        return TraversalCall(
            transform=lambda nodes: sorted([n.id for n in nodes]),
            sync_or_async=self,
            store=store,
            query=query,
            edges=edges,
            strategy=strategy,
        )

    def traverse(
        self,
        store: Adapter,
        query: str | None = None,
        edges: list[EdgeSpec] | EdgeFunction | None = None,
        strategy: Strategy | None = None,
    ) -> TraversalCall[list[Node]]:
        return TraversalCall(
            transform=lambda nodes: nodes,
            sync_or_async=self,
            store=store,
            query=query,
            edges=edges,
            strategy=strategy,
        )


class SyncTraversal(SyncOrAsync):
    async def _traverse(self, **kwargs):
        return traverse(**kwargs)


class AsyncTraversal(SyncOrAsync):
    async def _traverse(self, **kwargs):
        return await atraverse(**kwargs)


T = TypeVar("T")


def _pick(name: str, call: T | None, init: T | None) -> T:
    value = call or init
    if value is None:
        raise ValueError(f"{name} must be set in call or init")
    return value


class TraversalCall(Generic[T]):
    def __init__(
        self,
        transform: Callable[[list[Node]], T],
        sync_or_async: SyncOrAsync,
        store: Adapter,
        query: str | None = None,
        edges: list[EdgeSpec] | EdgeFunction | None = None,
        strategy: Strategy | None = None,
    ) -> None:
        self.transform = transform
        self.sync_or_async = sync_or_async
        self.store = store
        self.query = query
        self.edges = edges
        self.strategy = strategy

    async def __call__(
        self,
        query: str | None = None,
        edges: list[EdgeSpec] | EdgeFunction | None = None,
        strategy: Strategy | None = None,
        metadata_filter: dict[str, Any] | None = None,
        initial_root_ids: Sequence[str] = (),
        **kwargs: Any,
    ) -> T:
        strategy = _pick("strategy", strategy, self.strategy)
        strategy = dataclasses.replace(strategy, **kwargs)

        results = await self.sync_or_async._traverse(
            query=_pick("query", query, self.query),
            store=self.store,
            edges=_pick("edges", edges, self.edges),
            strategy=strategy,
            metadata_filter=metadata_filter,
            initial_root_ids=initial_root_ids,
        )
        return self.transform(results)


@pytest.fixture(scope="function", params=["sync", "async"])
def sync_or_async(request: pytest.FixtureRequest) -> SyncOrAsync:
    if request.param == "sync":
        return SyncTraversal()
    elif request.param == "async":
        return AsyncTraversal()
    else:
        raise ValueError(f"Unexpected value '{request.param}'")



"""Provide eager (breadth-first) traversal strategy."""

import dataclasses
from collections.abc import Iterable

from typing_extensions import override

from graph_retriever.strategies.base import NodeTracker, Strategy
from graph_retriever.types import Node


@dataclasses.dataclass
class Eager(Strategy):
    """
    Eager traversal strategy (breadth-first).

    This strategy selects all discovered nodes at each traversal step. It ensures
    breadth-first traversal by processing nodes layer by layer, which is useful for
    scenarios where all nodes at the current depth should be explored before proceeding
    to the next depth.

    Parameters
    ----------
    select_k :
        Maximum number of nodes to retrieve during traversal.
    start_k :
        Number of documents to fetch via similarity for starting the traversal.
        Added to any initial roots provided to the traversal.
    adjacent_k :
        Number of documents to fetch for each outgoing edge.
    max_depth :
        Maximum traversal depth. If `None`, there is no limit.
    k:
        Deprecated: Use `select_k` instead.
        Maximum number of nodes to select and return during traversal.
    """

    @override
    def iteration(self, nodes: Iterable[Node], tracker: NodeTracker) -> None:
        tracker.select_and_traverse(nodes)



"""Defines the `Node` class used during graph traversal."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from graph_retriever.edges import Edge


@dataclass
class Node:
    """
    Represents a node in the traversal graph.

    The [Node][graph_retriever.Node] class contains information about a document
    during graph traversal, including its depth, embedding, edges, and metadata.

    Parameters
    ----------
    id :
        The unique identifier of the document represented by this node.
    content :
        The content.
    depth :
        The depth (number of edges) through which this node was discovered. This
        depth may not reflect the true depth in the full graph if only a subset
        of edges is retrieved.
    embedding :
        The embedding vector of the document, used for similarity calculations.
    metadata :
        Metadata from the original document. This is a reference to the original
        document metadata and should not be modified directly. Any updates to
        metadata should be made to `extra_metadata`.
    extra_metadata :
        Additional metadata to override or augment the original document
        metadata during traversal.
    """

    id: str
    content: str
    depth: int
    similarity_score: float
    embedding: list[float]
    metadata: dict[str, Any] = field(default_factory=dict)

    incoming_edges: set[Edge] = field(default_factory=set)
    outgoing_edges: set[Edge] = field(default_factory=set)

    extra_metadata: dict[str, Any] = field(default_factory=dict)



from typing import Any

import pytest
from graph_retriever import Node
from graph_retriever.edges import Edges, IdEdge, MetadataEdge, MetadataEdgeFunction


def mk_node(metadata: dict[str, Any]) -> Node:
    return Node(
        id="id",
        content="testing",
        metadata=metadata,
        depth=0,
        embedding=[],
        similarity_score=0.2,
    )


def test_initialization():
    edge_function = MetadataEdgeFunction([("a", "a"), ("b", "c"), ("b", "b")])
    assert edge_function.edges == [("a", "a"), ("b", "c"), ("b", "b")]


def test_edge_function():
    edge_function = MetadataEdgeFunction([("href", "url")])
    assert edge_function(mk_node({"href": "a", "url": "b"})) == Edges(
        {MetadataEdge("url", "b")},
        {MetadataEdge("url", "a")},
    )

    assert edge_function(mk_node({"href": ["a", "c"], "url": "b"})) == Edges(
        {MetadataEdge("url", "b")},
        {MetadataEdge("url", "a"), MetadataEdge("url", "c")},
    )

    assert edge_function(mk_node({"href": ["a", "c"], "url": ["b", "d"]})) == Edges(
        {MetadataEdge("url", "b"), MetadataEdge("url", "d")},
        {MetadataEdge("url", "a"), MetadataEdge("url", "c")},
    )


def test_nested_edge():
    edge_function = MetadataEdgeFunction([("a.b", "b.c")])
    assert edge_function(mk_node({"a": {"b": 5}, "b": {"c": 7}})) == Edges(
        {MetadataEdge("b.c", 7)},
        {MetadataEdge("b.c", 5)},
    )


def test_link_to_id():
    edge_function = MetadataEdgeFunction([("mentions", "$id")])
    result = edge_function(mk_node({"mentions": ["a", "c"]}))

    assert result.incoming == {IdEdge("id")}
    assert result.outgoing == {IdEdge("a"), IdEdge("c")}


def test_link_from_id():
    edge_function = MetadataEdgeFunction([("$id", "mentions")])
    result = edge_function(mk_node({"mentions": ["a", "c"]}))

    assert result.incoming == {
        MetadataEdge("mentions", "a"),
        MetadataEdge("mentions", "c"),
    }
    assert result.outgoing == {MetadataEdge("mentions", "id")}


def test_unsupported_values():
    edge_function = MetadataEdgeFunction([("href", "url")])

    # Unsupported value
    with pytest.warns(UserWarning, match=r"Unsupported value .* in 'href'"):
        edge_function(mk_node({"href": None}))

    # Unsupported item value
    with pytest.warns(UserWarning, match=r"Unsupported item value .* in 'href'"):
        edge_function(mk_node({"href": [None]}))



"""
Specification and implementation of edges functions.

These are responsible for extracting edges from nodes and expressing them in way
that the adapters can implement.
"""

from ._base import Edge, EdgeFunction, Edges, IdEdge, MetadataEdge
from .metadata import EdgeSpec, Id, MetadataEdgeFunction

__all__ = [
    "Edge",
    "MetadataEdge",
    "IdEdge",
    "Edges",
    "EdgeFunction",
    "EdgeSpec",
    "Id",
    "MetadataEdgeFunction",
]



"""
Provides retrieval functions combining vector and graph traversal.

The main methods are [`traverse`][graph_retriever.traverse] and
[`atraverse`][graph_retriever.atraverse] which provide synchronous and
asynchronous traversals.
"""

from .content import Content
from .traversal import atraverse, traverse
from .types import Node

__all__ = [
    "Content",
    "Node",
    "traverse",
    "atraverse",
]



import asyncio
from collections.abc import AsyncIterator

import pytest
from graph_retriever.utils.merge import amerge


async def async_generator(
    values: list[int], event: asyncio.Event | None = None
) -> AsyncIterator[int]:
    """Helper async generator that yields values with an optional delay."""
    for value in values:
        if event:
            await event.wait()
            event.clear()
        yield value


async def test_amerge_basic():
    """Test merging two basic async iterators."""
    gen1 = async_generator([1, 3, 5])
    gen2 = async_generator([2, 4, 6])

    result = [val async for val in amerge(gen1, gen2)]
    assert sorted(result) == [1, 2, 3, 4, 5, 6]


async def test_merge_empty_iterators():
    """Test merging when one of the iterators is empty."""
    gen1 = async_generator([])
    gen2 = async_generator([1, 2, 3])

    result = [val async for val in amerge(gen1, gen2)]
    assert result == [1, 2, 3]  # Should return only the non-empty iterator's items


async def test_merge_all_empty():
    """Test merging when all iterators are empty."""
    gen1 = async_generator([])
    gen2 = async_generator([])

    result = [val async for val in amerge(gen1, gen2)]
    assert result == []  # Should return an empty list


async def test_merge_large_iterators():
    """Test merging large iterators."""
    gen1 = async_generator(range(100))
    gen2 = async_generator(range(100, 200))

    result = [val async for val in amerge(gen1, gen2)]
    result.sort()
    assert result == list(range(200))  # Ensure all items are included


async def test_merge_unordered_iterators():
    """Ensure iterators are merged in order of availability, not sorting."""
    e1 = asyncio.Event()
    e2 = asyncio.Event()
    gen1 = async_generator([10, 30, 50], e1)
    gen2 = async_generator([20, 40], e2)

    it = amerge(gen1, gen2)
    e1.set()
    assert await anext(it) == 10
    e1.set()
    assert await anext(it) == 30
    e1.set()
    assert await anext(it) == 50
    e1.set()
    e2.set()
    assert await anext(it) == 20
    e2.set()
    assert await anext(it) == 40
    e1.set()
    e2.set()
    assert await anext(it, None) is None


async def test_merge_exception_handling():
    """Ensure that an exception in one iterator does not break everything."""

    async def faulty_generator():
        """Async generator that raises an exception mid-way."""
        yield 1
        yield 2
        raise ValueError("Test exception")
        yield 3  # Should never be reached

    gen1 = async_generator([10, 20, 30])
    gen2 = faulty_generator()

    with pytest.raises(ValueError, match="Test exception"):
        _result = [val async for val in amerge(gen1, gen2)]



import pytest
from graph_retriever.adapters.base import Adapter
from graph_retriever.strategies.scored import Scored
from graph_retriever.testing.adapter_tests import cosine_similarity_scores
from graph_retriever.types import Node

from tests.testing.adapters import ANIMALS_QUERY
from tests.testing.invoker import SyncOrAsync


def score_animals(node: Node) -> float:
    return (20 - len(node.id)) + ((ord(node.id[0]) - ord("a")) / 100)


async def test_animals_keywords(animals: Adapter, sync_or_async: SyncOrAsync):
    """Test traversing a bi-directional field with lists."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("keywords", "keywords")],
        strategy=Scored(scorer=score_animals, start_k=2),
    )

    # # start_k=2 => 2 closest matches to the query
    assert await traversal(max_depth=0) == [
        "fox",
        "mongoose",
    ]
    # # select_k=8, we start with 2 closest and choose 6 more with shortest names
    assert await traversal(select_k=8, max_depth=1) == [
        "cat",
        "coyote",
        "fox",
        "gazelle",
        "hyena",
        "jackal",
        "mongoose",
    ]
    # select_k=4, we start with 2 closest and choose 2 more with shortest names
    # (from "cat", "coyote", "gazelle", "hyena", "jackal")
    assert await traversal(select_k=4, max_depth=1) == [
        "cat",
        "fox",
        "hyena",
        "jackal",
    ]


async def test_animals_populates_metrics_and_order(
    animals: Adapter, sync_or_async: SyncOrAsync
):
    """Test that score and depth are populated and results are returned in order."""
    results = await sync_or_async.traverse(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("habitat", "habitat")],
        strategy=Scored(scorer=score_animals, start_k=2),
    )(select_k=8, max_depth=2)

    expected_scores = {
        "mongoose": 12.12,
        "bobcat": 14.01,
        "cobra": 15.02,
        "deer": 16.03,
        "elk": 17.04,
        "fox": 17.05,
    }
    expected_depths = {
        "mongoose": 0,
        "bobcat": 1,
        "cobra": 1,
        "deer": 1,
        "elk": 1,
        "fox": 0,
    }

    expected_similarity_scores = cosine_similarity_scores(
        animals, ANIMALS_QUERY, list(expected_depths.keys())
    )

    for n in results:
        assert n.extra_metadata["_similarity_score"] == pytest.approx(
            expected_similarity_scores[n.id]
        ), f"incorrect similarity score for {n.id}"
        assert n.extra_metadata["_score"] == expected_scores[n.id], (
            f"incorrect score for {n.id}"
        )
        assert n.extra_metadata["_depth"] == expected_depths[n.id], (
            f"incorrect depth for {n.id}"
        )

    expected_ids_in_order = sorted(
        expected_scores.keys(), key=lambda id: expected_scores[id], reverse=True
    )
    assert [n.id for n in results] == expected_ids_in_order, (
        "incorrect order of results"
    )



import pytest
from graph_retriever import Content
from graph_retriever.adapters import Adapter
from graph_retriever.adapters.in_memory import InMemory
from graph_retriever.edges import Edges, MetadataEdge
from graph_retriever.strategies import (
    Eager,
)
from graph_retriever.testing.adapter_tests import cosine_similarity_scores
from graph_retriever.testing.embeddings import (
    ParserEmbeddings,
    angular_2d_embedding,
    earth_embeddings,
)

from tests.testing.adapters import (
    ANIMALS_DEPTH_0_EXPECTED,
    ANIMALS_QUERY,
)
from tests.testing.invoker import SyncOrAsync


async def test_earth(sync_or_async: SyncOrAsync):
    embedding = earth_embeddings
    greetings = Content(
        id="greetings",
        content="Typical Greetings",
        embedding=embedding("Typical Greetings"),
        metadata={
            "incoming": "parent",
        },
    )

    doc1 = Content(
        id="doc1",
        content="Hello World",
        embedding=embedding("Hello World"),
        metadata={"outgoing": "parent", "keywords": ["greeting", "world"]},
    )

    doc2 = Content(
        id="doc2",
        content="Hello Earth",
        embedding=embedding("Hello Earth"),
        metadata={"outgoing": "parent", "keywords": ["greeting", "earth"]},
    )

    store = InMemory(embedding, [greetings, doc1, doc2])

    traversal = sync_or_async.traverse_sorted_ids(
        store=store,
        query="Earth",
        edges=[("outgoing", "incoming"), ("keywords", "keywords")],
        strategy=Eager(select_k=10),
    )

    assert await traversal(start_k=1, max_depth=0) == ["doc2"]
    assert await traversal(start_k=2, max_depth=0) == ["doc1", "doc2"]
    assert await traversal(start_k=1, max_depth=1) == ["doc1", "doc2", "greetings"]


async def test_animals_select_k(animals: Adapter, sync_or_async: SyncOrAsync):
    """Test traversing a bi-directional field with lists."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("keywords", "keywords")],
        strategy=Eager(),
    )
    assert len(await traversal(select_k=5)) == 5
    assert len(await traversal(select_k=3)) == 3


async def test_animals_keywords(animals: Adapter, sync_or_async: SyncOrAsync):
    """Test traversing a bi-directional field with lists."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("keywords", "keywords")],
        strategy=Eager(select_k=100, start_k=2),
    )

    assert await traversal(max_depth=0) == ANIMALS_DEPTH_0_EXPECTED
    assert await traversal(max_depth=1) == [
        "cat",
        "coyote",
        "fox",
        "gazelle",
        "hyena",
        "jackal",
        "mongoose",
    ]
    assert await traversal(max_depth=2) == [
        "alpaca",
        "bison",
        "cat",
        "coyote",
        "crow",
        "dog",
        "fox",
        "gazelle",
        "horse",
        "hyena",
        "jackal",
        "mongoose",
    ]


async def test_animals_habitat(animals: Adapter, sync_or_async: SyncOrAsync):
    """Test traversing a bi-directional field with singular values."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("habitat", "habitat")],
        strategy=Eager(select_k=100, start_k=2),
    )

    assert await traversal(max_depth=0) == ANIMALS_DEPTH_0_EXPECTED
    assert await traversal(max_depth=1) == [
        "bobcat",
        "cobra",
        "deer",
        "elk",
        "fox",
        "mongoose",
    ]
    assert await traversal(max_depth=2) == [
        "bobcat",
        "cobra",
        "deer",
        "elk",
        "fox",
        "mongoose",
    ]


async def test_animals_populates_metrics_and_order(
    animals: Adapter, sync_or_async: SyncOrAsync
):
    """Test that score and depth are populated and results are returned in order."""
    results = await sync_or_async.traverse(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("habitat", "habitat")],
        strategy=Eager(select_k=100, start_k=2, max_depth=2),
    )()

    expected_depths = {
        "mongoose": 0,
        "bobcat": 1,
        "cobra": 1,
        "deer": 1,
        "elk": 1,
        "fox": 0,
    }

    expected_similarity_scores = cosine_similarity_scores(
        animals, ANIMALS_QUERY, list(expected_depths.keys())
    )

    for n in results:
        assert n.extra_metadata["_similarity_score"] == pytest.approx(
            expected_similarity_scores[n.id]
        ), f"incorrect similarity score for {n.id}"
        assert n.extra_metadata["_depth"] == expected_depths[n.id], (
            f"incorrect depth for {n.id}"
        )

    expected_ids_in_order = sorted(
        expected_similarity_scores.keys(),
        key=lambda id: expected_similarity_scores[id],
        reverse=True,
    )
    assert [n.id for n in results] == expected_ids_in_order, (
        "incorrect order of results"
    )


async def test_animals_habitat_to_keywords(
    animals: Adapter, sync_or_async: SyncOrAsync
):
    """Test traversing a from a singular field (habitat) to collection (keywords)."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("habitat", "keywords")],
        strategy=Eager(select_k=10, start_k=2),
    )

    assert await traversal(max_depth=0) == ANIMALS_DEPTH_0_EXPECTED
    assert await traversal(max_depth=1) == ["bear", "bobcat", "fox", "mongoose"]
    assert await traversal(max_depth=2) == [
        "bear",
        "bobcat",
        "caribou",
        "fox",
        "mongoose",
    ]


async def test_animals_initial_roots(animals: Adapter, sync_or_async: SyncOrAsync):
    """Test traversing with initial root IDs."""
    traversal = sync_or_async.traverse_sorted_ids(
        store=animals,
        query=ANIMALS_QUERY,
        edges=[("keywords", "keywords")],
        strategy=Eager(select_k=10, start_k=0),
    )

    assert await traversal(initial_root_ids=["bobcat"], max_depth=0) == [
        "bobcat",
    ]
    assert await traversal(initial_root_ids=["bobcat"], max_depth=1) == [
        "bear",
        "bobcat",
    ]
    assert await traversal(initial_root_ids=["bobcat"], max_depth=2) == [
        "bear",
        "bobcat",
        "moose",
        "ostrich",
    ]
    assert await traversal(
        initial_root_ids=["bobcat", "cheetah"], select_k=20, max_depth=2
    ) == [
        "bear",
        "bobcat",
        "cassowary",
        "cheetah",
        "dingo",
        "eagle",
        "emu",
        "falcon",
        "hawk",
        "jaguar",
        "kangaroo",
        "leopard",
        "moose",
    ]


async def test_parsed(sync_or_async: SyncOrAsync):
    """
    This is a test of set of Documents to pre-populate,
    a graph vector store with entries placed in a certain way.

    Space of the entries (under Euclidean similarity):

                      A0    (*)
        ....        AL   AR       <....
        :
        :
        v              |  .           v
                       |   :
       TR              |   :
    T0   --------------x--------------   B0
       TL              |   :
                       |   :
                       |  .
                       | .
                       |
                    FL   FR
                      F0

    the query point is meant to be at (*).
    the A are bidirectionally with B
    the A are outgoing to T
    the A are incoming from F
    The links are like: L with L, 0 with 0 and R with R.
    """
    embedding = ParserEmbeddings(2)
    docs_a = [
        Content.new("AL", "[-1, 9]", embedding),
        Content.new("A0", "[0, 10]", embedding),
        Content.new("AR", "[1, 9]", embedding),
    ]
    docs_b = [
        Content.new("BL", "[9, 1]", [9.0, 1.0]),
        Content.new("B0", "[10, 0]", [10.0, 0.0]),
        Content.new("BR", "[9, -1]", [9.0, -1.0]),
    ]
    docs_f = [
        Content.new("FL", "[1, -9]", [1.0, -9.0]),
        Content.new("F0", "[0, -10]", [0.0, -10.0]),
        Content.new("FR", "[-1, -9]", [-1.0, -9.0]),
    ]
    docs_t = [
        Content.new("TL", "[-9, -1]", [-9.0, -1.0]),
        Content.new("T0", "[-10, 0]", [-10.0, 0.0]),
        Content.new("TR", "[-9, 1]", [-9.0, 1.0]),
    ]
    for doc_a, suffix in zip(docs_a, ["l", "0", "r"]):
        doc_a.metadata["tag"] = f"ab_{suffix}"
        doc_a.metadata["out"] = f"at_{suffix}"
        doc_a.metadata["in"] = f"af_{suffix}"
    for doc_b, suffix in zip(docs_b, ["l", "0", "r"]):
        doc_b.metadata["tag"] = f"ab_{suffix}"
    for doc_t, suffix in zip(docs_t, ["l", "0", "r"]):
        doc_t.metadata["in"] = f"at_{suffix}"
    for doc_f, suffix in zip(docs_f, ["l", "0", "r"]):
        doc_f.metadata["out"] = f"af_{suffix}"
    documents = docs_a + docs_b + docs_f + docs_t

    traversal = sync_or_async.traverse_sorted_ids(
        store=InMemory(embedding, documents),
        edges=[("out", "in"), ("tag", "tag")],
        query="[2, 10]",
        strategy=Eager(select_k=10, start_k=2),
    )

    assert await traversal(max_depth=0) == ["A0", "AR"]
    assert await traversal(max_depth=2) == ["A0", "AR", "B0", "BR", "T0", "TR"]


async def test_ids(sync_or_async: SyncOrAsync):
    embedding = angular_2d_embedding
    v0 = Content.new("v0", "-0.124", embedding)
    v1 = Content.new("v1", "+0.127", embedding, metadata={"mentions": ["v0"]})
    v2 = Content.new("v2", "+0.250", embedding, metadata={"mentions": ["v1", "v3"]})
    v3 = Content.new("v3", "+1.000", embedding)
    store = InMemory(embedding, [v0, v1, v2, v3])

    mentions_to_id = sync_or_async.traverse_sorted_ids(
        store=store,
        query="+0.249",
        strategy=Eager(start_k=1),
        edges=[("mentions", "$id")],
    )
    assert await mentions_to_id(max_depth=0) == ["v2"]
    assert await mentions_to_id(max_depth=1) == ["v1", "v2", "v3"]
    assert await mentions_to_id(max_depth=2) == ["v0", "v1", "v2", "v3"]

    id_to_mentions = sync_or_async.traverse_sorted_ids(
        store=store,
        query="-0.125",
        strategy=Eager(start_k=1),
        edges=[("$id", "mentions")],
    )
    assert await id_to_mentions(max_depth=0) == ["v0"]
    assert await id_to_mentions(max_depth=1) == ["v0", "v1"]
    assert await id_to_mentions(max_depth=2) == ["v0", "v1", "v2"]


async def test_edge_functions(sync_or_async: SyncOrAsync):
    embedding = angular_2d_embedding
    v0 = Content.new(
        "v0",
        "-0.124",
        embedding,
        metadata={"links": [("a", 5.0)], "incoming": ["a"]},
    )
    v1 = Content.new(
        "v1",
        "+1.000",
        embedding,
        metadata={"links": [("a", 6.0)], "incoming": ["a"]},
    )
    store = InMemory(embedding, [v0, v1])

    def link_function(node: Content) -> Edges:
        links = node.metadata.get("links", [])
        incoming = node.metadata.get("incoming", [])
        return Edges(
            incoming={MetadataEdge("incoming", v) for v in incoming},
            outgoing={MetadataEdge("incoming", v) for v, _weight in links},
        )

    traversal = sync_or_async.traverse_sorted_ids(
        store=store,
        query="-0.125",
        edges=link_function,
        strategy=Eager(start_k=1),
    )
    assert await traversal(max_depth=0) == ["v0"]
    assert await traversal(max_depth=1) == ["v0", "v1"]
